////////////////////////////////////////////////////////////////////////
// Class:       FilterSimPhotonTime
// Module Type: filter
// File:        FilterSimPhotonTime_module.cc
//
// Generated at Tue Jan 19 09:42:51 2016 by Wesley Ketchum using artmod
// from cetpkgsupport v1_10_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/exception.h"

#include <memory>

#include <string>
#include <vector>
#include <iostream>
#include <algorithm>

#include "lardataobj/Simulation/SimPhotons.h"

namespace simfilter {
  class FilterSimPhotonTime;
}

class simfilter::FilterSimPhotonTime : public art::EDFilter {
public:
  explicit FilterSimPhotonTime(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  FilterSimPhotonTime(FilterSimPhotonTime const &) = delete;
  FilterSimPhotonTime(FilterSimPhotonTime &&) = delete;
  FilterSimPhotonTime & operator = (FilterSimPhotonTime const &) = delete;
  FilterSimPhotonTime & operator = (FilterSimPhotonTime &&) = delete;

  // Required functions.
  bool filter(art::Event & e) override;

  // Selected optional functions.
  //void reconfigure(fhicl::ParameterSet const & p) ;

private:

  std::string                       const fSimPhotonsCollectionLabel;
  std::vector< std::vector<float> > const fTimeWindows;
  float                             const fMinTotalEnergy;
  float                             const fMinPhotonEnergy;
  bool                              const fDebug; 

  std::vector<float>                fSumEnergyArray;
  
  void CheckTimeWindows();
};


simfilter::FilterSimPhotonTime::FilterSimPhotonTime(fhicl::ParameterSet const & p)
  :
  fSimPhotonsCollectionLabel(p.get<std::string>("SimPhotonsCollectionLabel")),
  fTimeWindows(p.get< std::vector< std::vector<float> > >("TimeWindows")),
  fMinTotalEnergy(p.get<float>("MinTotalEnergy",0.0)),
  fMinPhotonEnergy(p.get<float>("MinPhotonEnergy",-1)),
  fDebug(p.get<bool>("Debug",false)),
  fSumEnergyArray(fTimeWindows.size())
{
  CheckTimeWindows();
}

void simfilter::FilterSimPhotonTime::CheckTimeWindows(){

  if(fDebug)
    std::cout << "\tFilterSimPhotonTime: TimeWindows size is "
	      << fTimeWindows.size() << std::endl;
  
  for(auto const& tw : fTimeWindows){


    if(tw.size()!=2)
      throw cet::exception("FilterSimPhotonTime::CheckTimeWindows")
	<< "Bad time window initialization: time window has wrong size (not 2)."
	<< std::endl;
    
    if(fDebug)
      std::cout << "\t\tTimeWindow "
		<< "[" << tw[0] << "," << tw[1] << "]"
		<< std::endl;
    
    if(tw[0]>tw[1])
      throw cet::exception("FilterSimPhotonTime::CheckTimeWindows")
	<< "Bad time window initialization: tw[0]>tw[1]. Reverse the order!"
	<< std::endl;
    
    
  }

}

bool simfilter::FilterSimPhotonTime::filter(art::Event & e)
{

  art::Handle< std::vector<sim::SimPhotons> > simPhotonsHandle;
  e.getByLabel(fSimPhotonsCollectionLabel,simPhotonsHandle);
  std::vector<sim::SimPhotons> const& simPhotonsCollection(*simPhotonsHandle);

  std::fill(fSumEnergyArray.begin(),fSumEnergyArray.end(),0.0);
  for(auto const& simphotons : simPhotonsCollection){
    if(fDebug)
      std::cout << "\tFilterSimPhotonTime: Processing simphotons channel "
		<< simphotons.OpChannel()
		<< std::endl;

    for(auto const& photon : simphotons)
      for(size_t i_tw=0; i_tw<fTimeWindows.size(); ++i_tw){
	auto const& tw(fTimeWindows[i_tw]);
	if(photon.Time>=tw[0] && photon.Time<=tw[1] && photon.Energy>fMinPhotonEnergy){
	  
	  if(fDebug)
	    std::cout << "\t\tPhoton with time " << photon.Time << " detected. "
		      << "Energy is  " << photon.Energy << "."
		      << std::endl;
	  
	  fSumEnergyArray[i_tw] += photon.Energy;

	  if(fDebug)
	    std::cout << "\t\tTotal energy in this window (" << i_tw << ") is now " << fSumEnergyArray[i_tw] << std::endl;

	  if(fSumEnergyArray[i_tw] > fMinTotalEnergy) return true;
	}
      }
  }

  if(fDebug){
    std::cout << "\tFilterSimPhotonTime: Final total energies are below min of " << fMinTotalEnergy << ":" << std::endl;
    for(size_t i_tw=0; i_tw<fTimeWindows.size(); ++i_tw){
      std::cout << "\t\tTimeWindow "
		<< "[" << fTimeWindows[i_tw][0] << "," << fTimeWindows[i_tw][1] << "]: "
		<< fSumEnergyArray[i_tw]
		<< std::endl;
    }
  }
  
  //if(sum_energy > fMinTotalEnergy) return true;

  return false;
}

//void simfilter::FilterSimPhotonTime::reconfigure(fhicl::ParameterSet const & p)
//{
  // Implementation of optional member function here.
//}

DEFINE_ART_MODULE(simfilter::FilterSimPhotonTime)
